;/------------------------------------\
;| Thorium Sprite Image (TSI) Include |
;|                                    |
;| Version 0.02 Beta                  |
;| Last Edit: 10.14.2009              |
;|                                    |
;| PureBasic 4.40B5                   |
;\------------------------------------/

EnableExplicit

;---------------------------------------------------------
;Compiler Constants
;
;Here you can choose which functionalities you want to use,
;so only procedures you use will be compiled.

;Set to #True if you want to decode TSI's into PB images.
#Tsi_Compile_PbImageDecoder  = #True

;Set to #True if you want to decode TSI's into PB sprites.
#Tsi_Compile_PbSpriteDecoder = #True

;Set to #True if you want to encode PB images to TSI's.
#Tsi_Compile_PbImageEncoder  = #True

;Set to #True if you want to encode PB sprites to TSI's.
#Tsi_Compile_PbSpriteEncoder = #True

;---------------------------------------------------------
;Library Imports
;
;All needed imports of static libs and DLL's.
;If you allready have imported one of this librarys in your
;code, just comment it out here.

ImportC "zlib.lib"
  compress2(*dest, *destLen, *source, sourceLen.i, level.i)
  uncompress(*dest, *destLen, *source, sourceLen)
EndImport

;---------------------------------------------------------
;TSI Public Constants
;
;You can use this constants as parameters for TSI procedures
;or to check error codes.
;
;Do not change anything from here!

#Tsi_Filter_None    = 0
#Tsi_Filter_Sub     = 1
#Tsi_Filter_Up      = 2
#Tsi_Filter_Average = 3

#Tsi_Transformation_None         = 0
#Tsi_Transformation_AlphaSort    = 1
#Tsi_Transformation_ChannelsSort = 2

#Tsi_Compression_None = 0
#Tsi_Compression_Zlib = 1
#Tsi_Compression_Thor = 2

Enumeration 1
  #Tsi_Error_AllocMem                 ;Error on allocating memory.
  #Tsi_Error_CreateFile               ;Can't create file.
  #Tsi_Error_NoAccess                 ;Can't access file.
  #Tsi_Error_ReadFile                 ;Error on reading data from file.
  #Tsi_Error_FileToSmall              ;File size is To small, file can't be a TSI.
  #Tsi_Error_WrongSignature           ;Signature is not matching TSI signature, file isn't a TSI.
  #Tsi_Error_NotSupportedVersion      ;Not supported format version.
  #Tsi_Error_DecompressZlib           ;Error on decompressing ZLib compressed TSI.
  #Tsi_Error_DecompressThor           ;Error on decompressing Thorium compressed TSI.
  #Tsi_Error_CompressZlib             ;Error on compressing image data with ZLib.
  #Tsi_Error_CompressThor             ;Error on compressing image data with Thorium.  
  #Tsi_Error_PbCreateImage            ;Error on creating PB image.
  #Tsi_Error_PbImageAccess            ;Can't access PB image.
  #Tsi_Error_PbImageBufferAccess      ;Can't access PB image buffer.
  #Tsi_Error_PbImageWrongPixelFormat  ;PB image has wrong pixel format.
  #Tsi_Error_PbCreateSprite           ;Error on creating PB sprite.
  #Tsi_Error_PbSpriteAccess           ;Can't access PB sprite.
  #Tsi_Error_PbSpriteBufferAccess     ;Can't access PB sprite buffer.
  #Tsi_Error_PbSpriteWrongPixelFormat ;PB sprite has wrong pixel format.
EndEnumeration

;---------------------------------------------------------
;TSI Public Global Variables
;
;You can use this variables to get informations about the
;TSI procedures.
;
;Do not change anything from here!

;Holds the error code of the last occurred error.
;You can check the meaning of the code with the #Tsi_Error_
;constants.
Global Tsi_LastError.i

;---------------------------------------------------------
;TSI Private Constants
;
;Constants used internaly by the TSI procedures.
;
;Do not change anything from here!

#Tsi_PB_PixelFormat_32Bits_BGR_ReversedY = #PB_PixelFormat_32Bits_BGR | #PB_PixelFormat_ReversedY
#Tsi_PB_PixelFormat_32Bits_RGB_ReversedY = #PB_PixelFormat_32Bits_RGB | #PB_PixelFormat_ReversedY

#Tsi_Signature = $6D497053726F6854

#Tsi_FormatVersion = 1

#Tsi_Flag_CompressedZlib = %00000001
#Tsi_Flag_CompressedThor = %00000010
#Tsi_Flag_AlphaSort      = %00000100
#Tsi_Flag_ChannelsSort   = %00001000
#Tsi_Flag_SubFilter      = %00010000
#Tsi_Flag_UpFilter       = %00100000
#Tsi_Flag_AverageFilter  = %00110000

;---------------------------------------------------------
;TSI Private Structures
;
;Structures used internaly by the TSI procedures.
;
;Do not change anything from here!

Structure Tsi_Header
  Signature.q
  FormatVersion.a
  Flags.a
  Width.u
  Height.u
EndStructure

Structure Tsi_CompressionHeader
  UncompSize.l
  CompSize.l
EndStructure

Structure Tsi_Pixel_BGR
  Blue.a
  Green.a
  Red.a
EndStructure

Structure Tsi_Pixel_BGRA
  Blue.a
  Green.a
  Red.a
  Alpha.a
EndStructure

Structure Tsi_Pixel_Channel
  Channel.a
EndStructure

;---------------------------------------------------------
;TSI Macros
;
;Macros used by the TSI procedures.
;
;Do not change anything from here!

Macro Tsi_BGRA2RGBA(Pixel)
  (Pixel & $FF00FF00) | (Pixel & $FF) << 16 | ((Pixel >> 16) & $FF)
EndMacro

;---------------------------------------------------------
;TSI Decoder, Shared Procedures
;
;Do not change anything from here!

CompilerIf #Tsi_Compile_PbImageDecoder = #True Or #Tsi_Compile_PbSpriteDecoder = #True

;Loads a TSI from a file to memory and returns the pointer to it.
;File size will be returned in *TsiSize.
;This procedure does not decode the TSI, it just loads the file unmodified to memory.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_LoadTsiFile(FileName.s, *TsiSize.Integer)

  Protected.i File, ReadBytes
  Protected *FileBuffer
  
  File = ReadFile(#PB_Any, FileName)
  If File = 0
    Tsi_LastError = #Tsi_Error_NoAccess
    Debug "Tsi_LoadTsiFile: Can't access file: " + FileName
    ProcedureReturn 0
  EndIf
  
  *TsiSize\i = Lof(File)
  If *TsiSize\i < SizeOf(Tsi_Header) + 4
    Tsi_LastError = #Tsi_Error_FileToSmall
    CloseFile(File)
    Debug "Tsi_LoadTsiFile: File size to small, file can't be a TSI."
    ProcedureReturn 0
  EndIf
  
  *FileBuffer = AllocateMemory(*TsiSize\i)
  If *FileBuffer = 0
    Tsi_LastError = #Tsi_Error_AllocMem
    Debug "Tsi_LoadTsiFile: Error on allocating memory."
    ProcedureReturn 0
  EndIf

  ReadBytes = ReadData(File, *FileBuffer, *TsiSize\i)
  CloseFile(File)
  
  If ReadBytes <> *TsiSize\i
    FreeMemory(*FileBuffer)
    Tsi_LastError = #Tsi_Error_ReadFile
    Debug "Tsi_LoadTsiFile: Error on reading data from file."
    ProcedureReturn 0
  EndIf

  ProcedureReturn *FileBuffer

EndProcedure

;Undos the sort alpha transformation.
Procedure Tsi_UnSortAlpha(*ImageData, Size.i)

  Protected.i ByteSize
  Protected *Buffer

  ByteSize = Size * 4
  *Buffer = AllocateMemory(ByteSize)
  CopyMemory(*ImageData, *Buffer, ByteSize)

  CompilerSelect #PB_Compiler_Processor
  
    CompilerCase #PB_Processor_x86
      
      !push eax
      !push ecx
      !push edi
      !push esi
      !push ebx

      !mov ecx,[p.v_Size+20]
      !mov edi,[p.p_ImageData+20]
      !mov esi,[p.p_Buffer+20]
      !mov ebx,esi
      !add ebx,ecx
      
      !align 4
      !Tsi_UnSortAlpha_LoopStart:
        
        !mov al,[ebx]
        !mov [edi],al
        !inc ebx
        !inc edi
        
        !mov al,[ebx]
        !mov [edi],al
        !inc ebx
        !inc edi

        !mov al,[ebx]
        !mov [edi],al
        !inc ebx
        !inc edi

        !mov al,[esi]
        !mov [edi],al
        !inc esi
        !inc edi
        
      !dec ecx
      !cmp ecx,0
      !jne Tsi_UnSortAlpha_LoopStart

      !pop ebx
      !pop esi
      !pop edi
      !pop ecx
      !pop eax

    CompilerCase #PB_Processor_x64
    
      !push rax
      !push rcx
      !push rdi
      !push rsi
      !push rbx

      !mov rcx,[p.v_Size+40]
      !mov rdi,[p.p_ImageData+40]
      !mov rsi,[p.p_Buffer+40]
      !mov rbx,rsi
      !add rbx,rcx
      
      !align 8
      !Tsi_UnSortAlpha_LoopStart:
        
        !mov al,[rbx]
        !mov [rdi],al
        !inc rbx
        !inc rdi
        
        !mov al,[rbx]
        !mov [rdi],al
        !inc rbx
        !inc rdi

        !mov al,[rbx]
        !mov [rdi],al
        !inc rbx
        !inc rdi

        !mov al,[rsi]
        !mov [rdi],al
        !inc rsi
        !inc rdi
        
      !dec rcx
      !cmp rcx,0
      !jne Tsi_UnSortAlpha_LoopStart

      !pop rbx
      !pop rsi
      !pop rdi
      !pop rcx
      !pop rax

    CompilerDefault
    
      Protected.i i
      Protected *PixelPos.Tsi_Pixel_BGRA
      Protected *AlphaPos.Tsi_Pixel_Channel
      Protected *ColorPos.Tsi_Pixel_BGR
      
      *PixelPos = *ImageData
      *AlphaPos = *Buffer
      *ColorPos = *Buffer + Size
       
      For i = 1 To Size
        
        *PixelPos\Alpha = *AlphaPos\Channel
        *AlphaPos + 1
        
        *PixelPos\Blue  = *ColorPos\Blue 
        *PixelPos\Green = *ColorPos\Green
        *PixelPos\Red   = *ColorPos\Red
        *ColorPos + 3
        
        *PixelPos + 4
    
      Next

  CompilerEndSelect

  FreeMemory(*Buffer)

EndProcedure

;Undos the sort channels transformation.
Procedure Tsi_UnSortChannels(*ImageData, Size.i)

  Protected.i ByteSize
  Protected *Buffer

  ByteSize = Size * 4
  *Buffer = AllocateMemory(ByteSize)
  CopyMemory(*ImageData, *Buffer, ByteSize)

  CompilerSelect #PB_Compiler_Processor
  
    CompilerCase #PB_Processor_x86
      
      !push eax
      !push ecx
      !push edi
      !push edx
      !push ebx
      !push ebp
      !push esi
      
      !mov ecx,[p.v_Size+28]
      !mov edi,[p.p_ImageData+28]
      !mov edx,[p.p_Buffer+28]
      !mov ebx,edx
      !add ebx,ecx
      !mov ebp,ebx
      !add ebp,ecx
      !mov esi,ebp
      !add esi,ecx

      !align 4
      !Tsi_UnSortChannels_LoopStart:

        !mov al,[edx]
        !mov [edi],al
        !inc edx
        !inc edi
        
        !mov al,[ebx]
        !mov [edi],al
        !inc ebx
        !inc edi
        
        !mov al,[ebp]
        !mov [edi],al
        !inc ebp
        !inc edi
        
        !mov al,[esi]
        !mov [edi],al
        !inc esi
        !inc edi
      
      !dec ecx
      !cmp ecx,0
      !jne Tsi_UnSortChannels_LoopStart
      
      !pop esi
      !pop ebp
      !pop ebx
      !pop edx
      !pop edi
      !pop ecx
      !pop eax

    CompilerCase #PB_Processor_x64

      !push rax
      !push rcx
      !push rdi
      !push rdx
      !push rbx
      !push rbp
      !push rsi
      
      !mov rcx,[p.v_Size+56]
      !mov rdi,[p.p_ImageData+56]
      !mov rdx,[p.p_Buffer+56]
      !mov rbx,rdx
      !add rbx,rcx
      !mov rbp,rbx
      !add rbp,rcx
      !mov rsi,rbp
      !add rsi,rcx

      !align 8
      !Tsi_UnSortChannels_LoopStart:

        !mov al,[rdx]
        !mov [rdi],al
        !inc rdx
        !inc rdi
        
        !mov al,[rbx]
        !mov [rdi],al
        !inc rbx
        !inc rdi
        
        !mov al,[rbp]
        !mov [rdi],al
        !inc rbp
        !inc rdi
        
        !mov al,[rsi]
        !mov [rdi],al
        !inc rsi
        !inc rdi
      
      !dec rcx
      !cmp rcx,0
      !jne Tsi_UnSortChannels_LoopStart
      
      !pop rsi
      !pop rbp
      !pop rbx
      !pop rdx
      !pop rdi
      !pop rcx
      !pop rax

    CompilerDefault
      
      Protected.i i
      Protected *PixelPos.Tsi_Pixel_BGRA
      Protected *BluePos.Tsi_Pixel_Channel
      Protected *GreenPos.Tsi_Pixel_Channel
      Protected *RedPos.Tsi_Pixel_Channel  
      Protected *AlphaPos.Tsi_Pixel_Channel

      *PixelPos = *ImageData
      
      *BluePos  = *Buffer
      *GreenPos = *BluePos + Size
      *RedPos   = *GreenPos + Size
      *AlphaPos = *RedPos + Size
      
      For i = 1 To Size
        
        *PixelPos\Blue  = *BluePos\Channel
        *BluePos + 1
        
        *PixelPos\Green = *GreenPos\Channel
        *GreenPos + 1
    
        *PixelPos\Red   = *RedPos\Channel
        *RedPos + 1
    
        *PixelPos\Alpha = *AlphaPos\Channel
        *AlphaPos + 1
        
        *PixelPos + 4
    
      Next
  
  CompilerEndSelect
  
  FreeMemory(*Buffer)

EndProcedure

;Undos the sub filter.
Procedure Tsi_UnFilterSub(*ImageData, Width.i, Height.i)

  CompilerSelect #PB_Compiler_Processor
  
    CompilerCase #PB_Processor_x86

      !push eax
      !push ecx
      !push edx
      !push ebx
      !push esi
      !push edi
      
      !mov esi,[p.p_ImageData+24]
      !mov edi,esi
      !add edi,4
      
      !mov ebx,[p.v_Width+24]
      !dec ebx
      
      !mov edx,[p.v_Height+24]
      
      !align 4
      !Tsi_UnFilterSub_Y_LoopStart:
      
        !mov ecx,ebx
        
        !align 4
        !Tsi_UnFilterSub_X_LoopStart:
      
          !mov al,[edi]
          !add al,[esi]
          !mov [edi],al
          !inc edi
          !inc esi

          !mov al,[edi]
          !add al,[esi]
          !mov [edi],al
          !inc edi
          !inc esi

          !mov al,[edi]
          !add al,[esi]
          !mov [edi],al
          !inc edi
          !inc esi

          !mov al,[edi]
          !add al,[esi]
          !mov [edi],al
          !inc edi
          !inc esi

          !dec ecx

        !cmp ecx,0
        !jne Tsi_UnFilterSub_X_LoopStart

        !add edi,4
        !add esi,4
        !dec edx

      !cmp edx,0
      !jne Tsi_UnFilterSub_Y_LoopStart

      !pop edi
      !pop esi
      !pop ebx
      !pop edx
      !pop ecx
      !pop eax
      
    CompilerCase #PB_Processor_x64

      !push rax
      !push rcx
      !push rdx
      !push rsi
      !push rdi
      !push r10

      !mov rsi,[p.p_ImageData+48]
      !mov rdi,rsi
      !add rdi,4
      
      !mov r10,[p.v_Width+48]
      !dec r10
      
      !mov rdx,[p.v_Height+48]
      
      !align 8
      !Tsi_UnFilterSub_Y_LoopStart:
      
        !mov rcx,r10
        
        !align 8
        !Tsi_UnFilterSub_X_LoopStart:
      
          !mov al,[rdi]
          !add al,[rsi]
          !mov [rdi],al
          !inc rdi
          !inc rsi

          !mov al,[rdi]
          !add al,[rsi]
          !mov [rdi],al
          !inc rdi
          !inc rsi

          !mov al,[rdi]
          !add al,[rsi]
          !mov [rdi],al
          !inc rdi
          !inc rsi

          !mov al,[rdi]
          !add al,[rsi]
          !mov [rdi],al
          !inc rdi
          !inc rsi

          !dec rcx

        !cmp rcx,0
        !jne Tsi_UnFilterSub_X_LoopStart

        !add rdi,4
        !add rsi,4
        !dec rdx

      !cmp rdx,0
      !jne Tsi_UnFilterSub_Y_LoopStart
      
      !pop r10
      !pop rdi
      !pop rsi
      !pop rdx
      !pop rcx
      !pop rax

    CompilerDefault

      Protected.i X, Y
      Protected *ActualChannel.Tsi_Pixel_Channel
      Protected *PriorChannel.Tsi_Pixel_Channel
    
      *PriorChannel  = *ImageData
      *ActualChannel = *ImageData + 4
    
      Width - 1
      
      For Y = 1 To Height
      
        For X = 1 To Width
        
          *ActualChannel\Channel = *ActualChannel\Channel + *PriorChannel\Channel
          *ActualChannel + 1
          *PriorChannel + 1
          
          *ActualChannel\Channel = *ActualChannel\Channel + *PriorChannel\Channel
          *ActualChannel + 1
          *PriorChannel + 1
    
          *ActualChannel\Channel = *ActualChannel\Channel + *PriorChannel\Channel
          *ActualChannel + 1
          *PriorChannel + 1
    
          *ActualChannel\Channel = *ActualChannel\Channel + *PriorChannel\Channel
          *ActualChannel + 1
          *PriorChannel + 1
        
        Next
        
        *PriorChannel + 4
        *ActualChannel + 4
        
      Next

  CompilerEndSelect
    
EndProcedure

;Undos the up filter.
Procedure Tsi_UnFilterUp(*ImageData, Width.i, Height.i)

  CompilerSelect #PB_Compiler_Processor
  
    CompilerCase #PB_Processor_x86

      !push eax
      !push ecx
      !push edx
      !push ebx
      !push esi
      !push edi

      !mov eax,[p.v_Height+24]
      !dec eax
      !mul dword[p.v_Width+24]
      !mov ecx,eax
      
      !mov edi,[p.p_ImageData+24]
      !mov esi,edi
      !mov edx,[p.v_Width+24]
      !shl edx,2
      !add edi,edx
      
      !align 4
      !Tsi_UnFilterUp_LoopStart:
      
        !mov al,[edi]
        !add al,[esi]
        !mov [edi],al
        !inc edi
        !inc esi

        !mov al,[edi]
        !add al,[esi]
        !mov [edi],al
        !inc edi
        !inc esi

        !mov al,[edi]
        !add al,[esi]
        !mov [edi],al
        !inc edi
        !inc esi

        !mov al,[edi]
        !add al,[esi]
        !mov [edi],al
        !inc edi
        !inc esi
        
        !dec ecx

      !cmp ecx,0
      !jne Tsi_UnFilterUp_LoopStart
    
      !pop edi
      !pop esi
      !pop ebx
      !pop edx
      !pop ecx
      !pop eax
    
    CompilerCase #PB_Processor_x64

      !push rax
      !push rcx
      !push rdx
      !push rsi
      !push rdi
      !push r8
      !push r9

      !mov rax,[p.v_Height+56]
      !dec rax
      !mul qword[p.v_Width+56]
      !mov rcx,rax

      !mov r8,[p.p_ImageData+56]
      !mov r9,r8
      !mov rdx,[p.v_Width+56]
      !shl rdx,2
      !add r8,rdx
      
      !align 8
      !Tsi_UnFilterUp_LoopStart:
      
        !mov al,[r8]
        !add al,[r9]
        !mov [r8],al
        !inc r8
        !inc r9

        !mov al,[r8]
        !add al,[r9]
        !mov [r8],al
        !inc r8
        !inc r9

        !mov al,[r8]
        !add al,[r9]
        !mov [r8],al
        !inc r8
        !inc r9

        !mov al,[r8]
        !add al,[r9]
        !mov [r8],al
        !inc r8
        !inc r9
        
        !dec rcx

      !cmp rcx,0
      !jne Tsi_UnFilterUp_LoopStart

      !pop r9
      !pop r8
      !pop rdi
      !pop rsi
      !pop rdx
      !pop rcx
      !pop rax
    
    CompilerDefault
    
      Protected.i X, ByteSize
      Protected *ActualChannel.Tsi_Pixel_Channel
      Protected *PriorChannel.Tsi_Pixel_Channel
      
      *PriorChannel  = *ImageData
      *ActualChannel = *ImageData + Width * 4
      
      Height - 1
      ByteSize = Width * Height * 4
      
      For X = 1 To ByteSize
        
        *ActualChannel\Channel = *ActualChannel\Channel + *PriorChannel\Channel
        *ActualChannel + 1
        *PriorChannel + 1
    
      Next

  CompilerEndSelect

EndProcedure

;Undos the average filter.
Procedure Tsi_UnFilterAverage(*ImageData, Width.i, Height.i)

  CompilerSelect #PB_Compiler_Processor
  
    CompilerCase #PB_Processor_x86

      !push eax
      !push ecx
      !push edx
      !push ebx
      !push esi
      !push edi

      !mov eax,[p.v_Width+24]
      !shl eax,2
      
      !mov edi,[p.p_ImageData+24]
      !mov esi,edi
      !mov ebp,edi
      !add edi,eax
      !add edi,4
      !add esi,4
      !add ebp,eax

      !mov ebx,[p.v_Width+24]
      !dec ebx
      
      !mov edx,[p.v_Height+24]
      !dec edx
      !push edx
      
      !align 4
      !Tsi_UnFilterAverage_Y_LoopStart:
      
        !mov ecx,ebx
        
        !align 4
        !Tsi_UnFilterAverage_X_LoopStart:

          !xor eax,eax
          !mov al,[esi]
          !xor edx,edx
          !mov dl,[ebp]
          !add eax,edx
          !shr eax,1
          !add al,[edi]
          !mov [edi],al
          !inc edi
          !inc esi
          !inc ebp

          !xor eax,eax
          !mov al,[esi]
          !xor edx,edx
          !mov dl,[ebp]
          !add eax,edx
          !shr eax,1
          !add al,[edi]
          !mov [edi],al
          !inc edi
          !inc esi
          !inc ebp

          !xor eax,eax
          !mov al,[esi]
          !xor edx,edx
          !mov dl,[ebp]
          !add eax,edx
          !shr eax,1
          !add al,[edi]
          !mov [edi],al
          !inc edi
          !inc esi
          !inc ebp

          !xor eax,eax
          !mov al,[esi]
          !xor edx,edx
          !mov dl,[ebp]
          !add eax,edx
          !shr eax,1
          !add al,[edi]
          !mov [edi],al
          !inc edi
          !inc esi
          !inc ebp

          !dec ecx

        !cmp ecx,0
        !jne Tsi_UnFilterAverage_X_LoopStart
        
        !add edi,4
        !add esi,4
        !add ebp,4
        !pop edx
        !dec edx
        !push edx

      !cmp edx,0
      !jne Tsi_UnFilterAverage_Y_LoopStart
      
      !pop edi
      !pop esi
      !pop ebx
      !pop edx
      !pop ecx
      !pop eax

    CompilerCase #PB_Processor_x64

      !push rax
      !push rcx
      !push rdx
      !push rbx
      !push rsi
      !push rdi
      !push r8
      !push r9
      !push r10
      !push r11

      
      !mov rax,[p.v_Width+80]
      !shl rax,2
      
      !mov r8,[p.p_ImageData+80]
      !mov r9,r8
      !mov r10,r8
      !add r8,rax
      !add r8,4
      !add r9,4
      !add r10,rax

      !mov rbx,[p.v_Width+80]
      !dec rbx
      
      !mov rdx,[p.v_Height+80]
      !dec rdx
      
      !align 8
      !Tsi_UnFilterAverage_Y_LoopStart:
      
        !mov rcx,rbx
        
        !align 8
        !Tsi_UnFilterAverage_X_LoopStart:

          !xor rax,rax
          !mov al,[r9]
          !xor r11,r11
          !mov r11l,[r10]
          !add rax,r11
          !shr rax,1
          !add al,[r8]
          !mov [r8],al
          !inc r8
          !inc r9
          !inc r10

          !xor rax,rax
          !mov al,[r9]
          !xor r11,r11
          !mov r11l,[r10]
          !add rax,r11
          !shr rax,1
          !add al,[r8]
          !mov [r8],al
          !inc r8
          !inc r9
          !inc r10

          !xor rax,rax
          !mov al,[r9]
          !xor r11,r11
          !mov r11l,[r10]
          !add rax,r11
          !shr rax,1
          !add al,[r8]
          !mov [r8],al
          !inc r8
          !inc r9
          !inc r10

          !xor rax,rax
          !mov al,[r9]
          !xor r11,r11
          !mov r11l,[r10]
          !add rax,r11
          !shr rax,1
          !add al,[r8]
          !mov [r8],al
          !inc r8
          !inc r9
          !inc r10

          !dec rcx

        !cmp rcx,0
        !jne Tsi_UnFilterAverage_X_LoopStart
        
        !add r8,4
        !add r9,4
        !add r10,4
        !dec rdx

      !cmp rdx,0
      !jne Tsi_UnFilterAverage_Y_LoopStart

      !pop r11
      !pop r10
      !pop r9
      !pop r8
      !pop rdi
      !pop rsi
      !pop rbx
      !pop rdx
      !pop rcx
      !pop rax

    CompilerDefault

      Protected.i X, Y, LineSize
      Protected *ActualChannel.Tsi_Pixel_Channel
      Protected *PriorChannelUp.Tsi_Pixel_Channel
      Protected *PriorChannelSub.Tsi_Pixel_Channel
    
      LineSize = Width * 4
      
      *PriorChannelUp  = *ImageData + 4
      *PriorChannelSub = *ImageData + LineSize
      *ActualChannel   = *ImageData + LineSize + 4
    
      Width - 1
      Height - 1
      
      For Y = 1 To Height
      
        For X = 1 To Width
        
          *ActualChannel\Channel = *ActualChannel\Channel + (*PriorChannelUp\Channel + *PriorChannelSub\Channel) / 2
          *ActualChannel + 1
          *PriorChannelUp + 1
          *PriorChannelSub + 1
    
          *ActualChannel\Channel = *ActualChannel\Channel + (*PriorChannelUp\Channel + *PriorChannelSub\Channel) / 2
          *ActualChannel + 1
          *PriorChannelUp + 1
          *PriorChannelSub + 1
    
          *ActualChannel\Channel = *ActualChannel\Channel + (*PriorChannelUp\Channel + *PriorChannelSub\Channel) / 2
          *ActualChannel + 1
          *PriorChannelUp + 1
          *PriorChannelSub + 1
    
          *ActualChannel\Channel = *ActualChannel\Channel + (*PriorChannelUp\Channel + *PriorChannelSub\Channel) / 2
          *ActualChannel + 1
          *PriorChannelUp + 1
          *PriorChannelSub + 1
          
        Next
        
        *PriorChannelSub + 4
        *PriorChannelUp + 4
        *ActualChannel + 4
        
      Next

  CompilerEndSelect
 
EndProcedure

;Decodes a TSI located in memory.
;Returns a pointer to the raw image data.
;Returns width and height of the image in *Width and *Height.
;Returns 0 if it fails, check Tsi_LastError for error code.
;Format of raw image data: pixel format is 32bit BGRA
;                          stored not reversed X and Y
;                          Y pitch = width * 4
Procedure.i Tsi_DecodeTsi(*Tsi, *Width.Integer, *Height.Integer)

  Protected.i ImageSizeByte
  Protected *DecompBuffer
  Protected *ImageDataStart
  Protected *Header.Tsi_Header
  Protected *CompHeader.Tsi_CompressionHeader

  *Header = *Tsi
  
  If *Header\Signature <> #Tsi_Signature
    Tsi_LastError = #Tsi_Error_WrongSignature
    Debug "Tsi_DecodeTsi: Signature is not matching TSI signature, file isn't a TSI."
    ProcedureReturn 0
  EndIf
  
  If *Header\FormatVersion <> #Tsi_FormatVersion
    Tsi_LastError = #Tsi_Error_NotSupportedVersion
    Debug "Tsi_DecodeTsi: Not supported format version: " + Str(*Header\FormatVersion)
    ProcedureReturn 0
  EndIf
  
  If *Header\Flags & #Tsi_Flag_CompressedZlib = #Tsi_Flag_CompressedZlib
    *CompHeader = *Header + SizeOf(Tsi_Header)
    *DecompBuffer = AllocateMemory(*CompHeader\UncompSize)
    If *DecompBuffer = 0
      Tsi_LastError = #Tsi_Error_AllocMem
      Debug "Tsi_DecodeTsi: Error on allocating memory."
      ProcedureReturn 0
    EndIf
    If uncompress(*DecompBuffer, @*CompHeader\UncompSize, *CompHeader + SizeOf(Tsi_CompressionHeader), *CompHeader\CompSize)
      FreeMemory(*DecompBuffer)
      Tsi_LastError = #Tsi_Error_DecompressZlib
      Debug "Tsi_DecodeTsi: Error on decompressing ZLib compressed TSI."
      ProcedureReturn 0 
    EndIf
    *ImageDataStart = *DecompBuffer
  ElseIf *Header\Flags & #Tsi_Flag_CompressedThor = #Tsi_Flag_CompressedThor
    Tsi_LastError = #Tsi_Error_DecompressThor
    Debug "Tsi_DecodeTsi: Image data is compressed with Thorium compression, not supported."
    ProcedureReturn 0
  Else
    ImageSizeByte = *Header\Width * *Header\Height * 4
    *ImageDataStart = AllocateMemory(ImageSizeByte)
    If *ImageDataStart = 0
      Tsi_LastError = #Tsi_Error_AllocMem
      Debug "Tsi_DecodeTsi: Error on allocating memory."
      ProcedureReturn 0
    EndIf
    CopyMemory(*Header + SizeOf(Tsi_Header), *ImageDataStart, ImageSizeByte)
  EndIf 

  If *Header\Flags & #Tsi_Flag_AlphaSort = #Tsi_Flag_AlphaSort
    Tsi_UnSortAlpha(*ImageDataStart, *Header\Width * *Header\Height)
  ElseIf *Header\Flags & #Tsi_Flag_ChannelsSort = #Tsi_Flag_ChannelsSort
    Tsi_UnSortChannels(*ImageDataStart, *Header\Width * *Header\Height)
  EndIf
  
  If *Header\Flags & #Tsi_Flag_AverageFilter = #Tsi_Flag_AverageFilter
    Tsi_UnFilterAverage(*ImageDataStart, *Header\Width, *Header\Height)
  ElseIf *Header\Flags & #Tsi_Flag_SubFilter = #Tsi_Flag_SubFilter
    Tsi_UnFilterSub(*ImageDataStart, *Header\Width, *Header\Height)
  ElseIf *Header\Flags & #Tsi_Flag_UpFilter = #Tsi_Flag_UpFilter
    Tsi_UnFilterUp(*ImageDataStart, *Header\Width, *Header\Height)
  EndIf

  *Width\i  = *Header\Width
  *Height\i = *Header\Height
  
  ProcedureReturn *ImageDataStart

EndProcedure

CompilerEndIf

;---------------------------------------------------------
;TSI Decoder, PB Image
;
;Do not change anything from here!

CompilerIf #Tsi_Compile_PbImageDecoder = #True

;Similar to PB's CatchImage. Decodes a TSI in memory to a PB image.
;Returns a PB image number.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_CatchTsi2Image(*Tsi)

  Protected.i X, Y, Image, Pitch, PixelFormat, SourceLineLen, DestY, Width, Height
  Protected *ImageBuffer
  Protected *RawImageData
  Protected *SourcePixel.Long
  Protected *DestPixel.Long

  *RawImageData = Tsi_DecodeTsi(*Tsi, @Width, @Height)
  If *RawImageData = 0
    ProcedureReturn 0
  EndIf
  
  Image = CreateImage(#PB_Any, Width, Height, 32)
  If Image = 0
    FreeMemory(*RawImageData)
    Tsi_LastError = #Tsi_Error_PbCreateImage
    Debug "Tsi_CatchTsi2Image: Error on creating PB image."
    ProcedureReturn 0
  EndIf

  If StartDrawing(ImageOutput(Image)) = 0
    FreeImage(Image)
    FreeMemory(*RawImageData)
    Tsi_LastError = #Tsi_Error_PbImageAccess
    Debug "Tsi_CatchTsi2Image: Can't access PB image."
    ProcedureReturn 0
  EndIf
  
  *ImageBuffer = DrawingBuffer()
  If *ImageBuffer = 0
    StopDrawing()
    FreeImage(Image)
    FreeMemory(*RawImageData)
    Tsi_LastError = #Tsi_Error_PbImageBufferAccess
    Debug "Tsi_CatchTsi2Image: Can't access PB image buffer."
    ProcedureReturn 0
  EndIf
  
  Pitch        = DrawingBufferPitch()
  PixelFormat  = DrawingBufferPixelFormat()
  
  SourceLineLen = Width * 4
  
  Select PixelFormat
  
    Case #PB_PixelFormat_32Bits_BGR
    
      If Pitch = SourceLineLen
        CopyMemory(*RawImageData, *ImageBuffer, SourceLineLen * Height)
      Else
        Height - 1
        For Y = 0 To Height
          CopyMemory(*RawImageData + (SourceLineLen * Y), *ImageBuffer + (Pitch * Y), SourceLineLen)
        Next
      EndIf
    
    Case #PB_PixelFormat_32Bits_RGB
      
      Height - 1
      Width - 1
      *SourcePixel = *RawImageData
      For Y = 0 To Height
        *DestPixel = *ImageBuffer + (Pitch * Y)
        For X = 0 To Width
          *DestPixel\l = Tsi_BGRA2RGBA(*SourcePixel\l)
          *SourcePixel + 4
          *DestPixel + 4
        Next
      Next
    
    Case #Tsi_PB_PixelFormat_32Bits_BGR_ReversedY
      
      Height - 1
      DestY = Height
      For Y = 0 To Height
        CopyMemory(*RawImageData + (SourceLineLen * Y), *ImageBuffer + (Pitch * DestY), SourceLineLen)
        DestY - 1
      Next
    
    Case #Tsi_PB_PixelFormat_32Bits_RGB_ReversedY

      Height - 1
      Width - 1
      *SourcePixel = *RawImageData
      DestY = Height
      For Y = 0 To Height
        *DestPixel = *ImageBuffer + (Pitch * DestY)
        For X = 0 To Width
          *DestPixel\l = Tsi_BGRA2RGBA(*SourcePixel\l)
          *SourcePixel + 4
          *DestPixel + 4
        Next
        DestY - 1
      Next
    
    Default
      StopDrawing()
      FreeImage(Image)
      FreeMemory(*RawImageData)
      Tsi_LastError = #Tsi_Error_PbImageWrongPixelFormat
      Debug "Tsi_CatchTsi2Image: PB image has wrong pixel format."
      ProcedureReturn 0
  
  EndSelect
  
  StopDrawing()
  FreeMemory(*RawImageData)
  
  ProcedureReturn Image

EndProcedure

;Similar to PB's LoadImage. Loads a TSI from a file and decodes it into a PB image.
;Returns a PB image number.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_LoadTsi2Image(FileName.s)

  Protected.i FileLength, Image
  Protected *TsiBuffer

  *TsiBuffer = Tsi_LoadTsiFile(FileName, @FileLength)
  If *TsiBuffer = 0
    ProcedureReturn 0
  EndIf
  
  Image = Tsi_CatchTsi2Image(*TsiBuffer)
  
  FreeMemory(*TsiBuffer)
  
  ProcedureReturn Image

EndProcedure

CompilerEndIf

;---------------------------------------------------------
;TSI Decoder, PB Sprite
;
;Do not change anything from here!

CompilerIf #Tsi_Compile_PbSpriteDecoder = #True

;Similar to PB's CatchSprite. Decodes a TSI in memory to a PB sprite.
;Returns a PB sprite number.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_CatchTsi2Sprite(*Tsi)

  Protected.i X, Y, Sprite, Pitch, PixelFormat, SourceLineLen, DestY, Width, Height
  Protected *SpriteBuffer
  Protected *RawImageData
  Protected *SourcePixel.Long
  Protected *DestPixel.Long

  *RawImageData = Tsi_DecodeTsi(*Tsi, @Width, @Height)
  If *RawImageData = 0
    ProcedureReturn 0
  EndIf
  
  Sprite = CreateSprite(#PB_Any, Width, Height, #PB_Sprite_Texture | #PB_Sprite_AlphaBlending)
  If Sprite = 0
    FreeMemory(*RawImageData)
    Tsi_LastError = #Tsi_Error_PbCreateSprite
    Debug "Tsi_CatchTsi2Sprite: Error on creating PB sprite."
    ProcedureReturn 0
  EndIf

  If StartDrawing(SpriteOutput(Sprite)) = 0
    FreeSprite(Sprite)
    FreeMemory(*RawImageData)
    Tsi_LastError = #Tsi_Error_PbSpriteAccess
    Debug "Tsi_CatchTsi2Sprite: Can't access PB sprite."
    ProcedureReturn 0
  EndIf
  
  *SpriteBuffer = DrawingBuffer()
  If *SpriteBuffer = 0
    StopDrawing()
    FreeImage(Sprite)
    FreeMemory(*RawImageData)
    Tsi_LastError = #Tsi_Error_PbSpriteBufferAccess
    Debug "Tsi_CatchTsi2Sprite: Can't access PB sprite buffer."
    ProcedureReturn 0
  EndIf
  
  Pitch        = DrawingBufferPitch()
  PixelFormat  = DrawingBufferPixelFormat()
  
  SourceLineLen = Width * 4
  
  Select PixelFormat
  
    Case #PB_PixelFormat_32Bits_BGR
    
      If Pitch = SourceLineLen
        CopyMemory(*RawImageData, *SpriteBuffer, SourceLineLen * Height)
      Else
        Height - 1
        For Y = 0 To Height
          CopyMemory(*RawImageData + (SourceLineLen * Y), *SpriteBuffer + (Pitch * Y), SourceLineLen)
        Next
      EndIf
    
    Case #PB_PixelFormat_32Bits_RGB
      
      Height - 1
      Width - 1
      *SourcePixel = *RawImageData
      For Y = 0 To Height
        *DestPixel = *SpriteBuffer + (Pitch * Y)
        For X = 0 To Width
          *DestPixel\l = Tsi_BGRA2RGBA(*SourcePixel\l)
          *SourcePixel + 4
          *DestPixel + 4
        Next
      Next
    
    Case #Tsi_PB_PixelFormat_32Bits_BGR_ReversedY
      
      Height - 1
      DestY = Height
      For Y = 0 To Height
        CopyMemory(*RawImageData + (SourceLineLen * Y), *SpriteBuffer + (Pitch * DestY), SourceLineLen)
        DestY - 1
      Next
    
    Case #Tsi_PB_PixelFormat_32Bits_RGB_ReversedY

      Height - 1
      Width - 1
      *SourcePixel = *RawImageData
      DestY = Height
      For Y = 0 To Height
        *DestPixel = *SpriteBuffer + (Pitch * DestY)
        For X = 0 To Width
          *DestPixel\l = Tsi_BGRA2RGBA(*SourcePixel\l)
          *SourcePixel + 4
          *DestPixel + 4
        Next
        DestY - 1
      Next
    
    Default
      StopDrawing()
      FreeSprite(Sprite)
      FreeMemory(*RawImageData)
      Tsi_LastError = #Tsi_Error_PbSpriteWrongPixelFormat
      Debug "Tsi_CatchTsi2Sprite: PB sprite has wrong pixel format."
      ProcedureReturn 0
  
  EndSelect
  
  StopDrawing()
  FreeMemory(*RawImageData)
  
  ProcedureReturn Sprite

EndProcedure

;Similar to PB's LoadSprite. Loads a TSI from a file and decodes it into a PB sprite.
;Returns a PB sprite number.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_LoadTsi2Sprite(FileName.s)

  Protected.i FileLength, Sprite
  Protected *TsiBuffer

  *TsiBuffer = Tsi_LoadTsiFile(FileName, @FileLength)
  If *TsiBuffer = 0
    ProcedureReturn 0
  EndIf
  
  Sprite = Tsi_CatchTsi2Sprite(*TsiBuffer)
  
  FreeMemory(*TsiBuffer)
  
  ProcedureReturn Sprite

EndProcedure

CompilerEndIf

;---------------------------------------------------------
;TSI Encoder, Shared Procedures
;
;Do not change anything from here!

CompilerIf #Tsi_Compile_PbImageEncoder = #True Or #Tsi_Compile_PbSpriteEncoder = #True

;Transforms the raw image data of a TSI, for bedder compression.
;Separates the alpha channel and moves it to the front of the image data.
Procedure Tsi_SortAlpha(*ImageData, Size.i)

  Protected.i i, ByteSize
  Protected *Buffer
  Protected *PixelPos.Tsi_Pixel_BGRA
  Protected *AlphaPos.Tsi_Pixel_Channel
  Protected *ColorPos.Tsi_Pixel_BGR

  ByteSize = Size * 4
  *Buffer = AllocateMemory(ByteSize)
  CopyMemory(*ImageData, *Buffer, ByteSize)
  
  *PixelPos = *Buffer
  *AlphaPos = *ImageData
  *ColorPos = *ImageData + Size
  
  For i = 1 To Size
    
    *AlphaPos\Channel = *PixelPos\Alpha
    *AlphaPos + 1

    *ColorPos\Blue  = *PixelPos\Blue
    *ColorPos\Green = *PixelPos\Green
    *ColorPos\Red   = *PixelPos\Red
    *ColorPos + 3
    
    *PixelPos + 4

  Next
  
  FreeMemory(*Buffer)

EndProcedure

;Transforms the raw image data of a TSI, for bedder compression.
;Separates all channels.
Procedure Tsi_SortChannels(*ImageData, Size.i)

  Protected.i i, ByteSize
  Protected *Buffer
  Protected *PixelPos.Tsi_Pixel_BGRA
  Protected *BluePos.Tsi_Pixel_Channel
  Protected *GreenPos.Tsi_Pixel_Channel
  Protected *RedPos.Tsi_Pixel_Channel  
  Protected *AlphaPos.Tsi_Pixel_Channel

  ByteSize = Size * 4
  *Buffer = AllocateMemory(ByteSize)
  CopyMemory(*ImageData, *Buffer, ByteSize)
  
  *PixelPos = *Buffer
  
  *BluePos  = *ImageData
  *GreenPos = *BluePos + Size
  *RedPos   = *GreenPos + Size
  *AlphaPos = *RedPos + Size
  
  For i = 1 To Size
    
    *BluePos\Channel  = *PixelPos\Blue
    *BluePos + 1

    *GreenPos\Channel = *PixelPos\Green
    *GreenPos + 1

    *RedPos\Channel   = *PixelPos\Red
    *RedPos + 1
    
    *AlphaPos\Channel = *PixelPos\Alpha
    *AlphaPos + 1
    
    *PixelPos + 4

  Next
  
  FreeMemory(*Buffer)

EndProcedure

;Filters the raw image data of a TSI, for bedder compression.
Procedure Tsi_FilterSub(*ImageData, Width.i, Height.i)

  Protected.i X, Y, ByteSize
  Protected *Buffer
  Protected *ActualChannel.Tsi_Pixel_Channel
  Protected *PriorChannel.Tsi_Pixel_Channel
  Protected *DestChannel.Tsi_Pixel_Channel

  ByteSize = Width * Height * 4
  *Buffer = AllocateMemory(ByteSize)
  CopyMemory(*ImageData, *Buffer, ByteSize)

  *DestChannel   = *ImageData + 4
  *PriorChannel  = *Buffer
  *ActualChannel = *Buffer + 4

  Width - 1
  
  For Y = 1 To Height
  
    For X = 1 To Width
    
      *DestChannel\Channel = *ActualChannel\Channel - *PriorChannel\Channel
      *DestChannel + 1
      *ActualChannel + 1
      *PriorChannel + 1
      
      *DestChannel\Channel = *ActualChannel\Channel - *PriorChannel\Channel
      *DestChannel + 1
      *ActualChannel + 1
      *PriorChannel + 1

      *DestChannel\Channel = *ActualChannel\Channel - *PriorChannel\Channel
      *DestChannel + 1
      *ActualChannel + 1
      *PriorChannel + 1

      *DestChannel\Channel = *ActualChannel\Channel - *PriorChannel\Channel
      *DestChannel + 1
      *ActualChannel + 1
      *PriorChannel + 1
    
    Next
    
    *PriorChannel + 4
    *ActualChannel + 4
    *DestChannel + 4
    
  Next
  
  FreeMemory(*Buffer)
  
EndProcedure

;Filters the raw image data of a TSI, for bedder compression.
Procedure Tsi_FilterUp(*ImageData, Width.i, Height.i)

  Protected.i X, LineSize, ByteSize
  Protected *Buffer
  Protected *ActualChannel.Tsi_Pixel_Channel
  Protected *PriorChannel.Tsi_Pixel_Channel
  Protected *DestChannel.Tsi_Pixel_Channel

  ByteSize = Width * Height * 4
  *Buffer = AllocateMemory(ByteSize)
  CopyMemory(*ImageData, *Buffer, ByteSize)

  LineSize = Width * 4
  
  *DestChannel   = *ImageData + LineSize
  *PriorChannel  = *Buffer
  *ActualChannel = *Buffer + LineSize
    
  Height - 1
  ByteSize = Width * Height * 4
  
  For X = 1 To ByteSize
    
    *DestChannel\Channel = *ActualChannel\Channel - *PriorChannel\Channel
    *DestChannel + 1
    *ActualChannel + 1
    *PriorChannel + 1

  Next
  
  FreeMemory(*Buffer)

EndProcedure

;Filters the raw image data of a TSI, for bedder compression.
Procedure Tsi_FilterAverage(*ImageData, Width.i, Height.i)

  Protected.i X, Y, LineSize, ByteSize
  Protected *Buffer
  Protected *ActualChannel.Tsi_Pixel_Channel
  Protected *PriorChannelUp.Tsi_Pixel_Channel
  Protected *PriorChannelSub.Tsi_Pixel_Channel
  Protected *DestChannel.Tsi_Pixel_Channel

  ByteSize = Width * Height * 4
  *Buffer = AllocateMemory(ByteSize)
  CopyMemory(*ImageData, *Buffer, ByteSize)

  LineSize = Width * 4
  
  *DestChannel     = *ImageData + LineSize + 4
  *PriorChannelUp  = *Buffer + 4
  *PriorChannelSub = *Buffer + LineSize
  *ActualChannel   = *Buffer + LineSize + 4

  Width - 1
  Height - 1
  
  For Y = 1 To Height
  
    For X = 1 To Width
    
      *DestChannel\Channel = *ActualChannel\Channel - (*PriorChannelUp\Channel + *PriorChannelSub\Channel) / 2
      *DestChannel + 1
      *ActualChannel + 1
      *PriorChannelUp + 1
      *PriorChannelSub + 1

      *DestChannel\Channel = *ActualChannel\Channel - (*PriorChannelUp\Channel + *PriorChannelSub\Channel) / 2
      *DestChannel + 1
      *ActualChannel + 1
      *PriorChannelUp + 1
      *PriorChannelSub + 1

      *DestChannel\Channel = *ActualChannel\Channel - (*PriorChannelUp\Channel + *PriorChannelSub\Channel) / 2
      *DestChannel + 1
      *ActualChannel + 1
      *PriorChannelUp + 1
      *PriorChannelSub + 1

      *DestChannel\Channel = *ActualChannel\Channel - (*PriorChannelUp\Channel + *PriorChannelSub\Channel) / 2
      *DestChannel + 1
      *ActualChannel + 1
      *PriorChannelUp + 1
      *PriorChannelSub + 1
      
    Next
    
    *PriorChannelSub + 4
    *PriorChannelUp + 4
    *ActualChannel + 4
    *DestChannel + 4
    
  Next

  FreeMemory(*Buffer)

EndProcedure

;Encodes raw image data to a TSI in memory.
;Returns a pointer to the TSI.
;Returns size of the TSI in *TsiSize.
;Returns 0 if it fails, check Tsi_LastError for error code.
;Format of raw image data: pixel format is 32bit BGRA
;                          stored not reversed X and Y
;                          Y pitch = width * 4
Procedure.i Tsi_EncodeTsi(*RawImageData, *TsiSize.Integer, Width.i, Height.i, Filter.i, Transformation.i, Compression.i, CompressionLevel.i)

  Protected.i ImageDataSize, CompBufferSize
  Protected *ImageDataStart
  Protected *TsiBuffer
  Protected *CompBuffer
  Protected *Header.Tsi_Header
  Protected *CompHeader.Tsi_CompressionHeader

  *TsiSize\i = 0
  ImageDataSize = Width * Height * 4

  If Compression <> #Tsi_Compression_None
    *TsiSize\i = SizeOf(Tsi_CompressionHeader)
    *ImageDataStart = SizeOf(Tsi_CompressionHeader)
  EndIf

  *TsiSize\i = *TsiSize\i + ImageDataSize + SizeOf(Tsi_Header)
  
  *TsiBuffer = AllocateMemory(*TsiSize\i)
  If *TsiBuffer = 0
    Tsi_LastError = #Tsi_Error_AllocMem
    Debug "Tsi_EncodeTsi: Error on allocating memory."
    ProcedureReturn 0
  EndIf
  
  *ImageDataStart = *ImageDataStart + *TsiBuffer + SizeOf(Tsi_Header)

  *Header = *TsiBuffer
  *Header\Signature     = #Tsi_Signature
  *Header\FormatVersion = #Tsi_FormatVersion
  *Header\Width         = Width
  *Header\Height        = Height

  CopyMemory(*RawImageData, *ImageDataStart, ImageDataSize)

  Select Filter
    Case #Tsi_Filter_Sub
      *Header\Flags = *Header\Flags | #Tsi_Flag_SubFilter
      Tsi_FilterSub(*ImageDataStart, Width, Height)      
    Case #Tsi_Filter_Up
      *Header\Flags = *Header\Flags | #Tsi_Flag_UpFilter
      Tsi_FilterUp(*ImageDataStart, Width, Height)
    Case #Tsi_Filter_Average
      *Header\Flags = *Header\Flags | #Tsi_Flag_SubFilter | #Tsi_Flag_UpFilter
      Tsi_FilterAverage(*ImageDataStart, Width, Height)
  EndSelect

  Select Transformation
    Case #Tsi_Transformation_AlphaSort
      *Header\Flags = *Header\Flags | #Tsi_Flag_AlphaSort
      Tsi_SortAlpha(*ImageDataStart, Width * Height)
    Case #Tsi_Transformation_ChannelsSort
      *Header\Flags = *Header\Flags | #Tsi_Flag_ChannelsSort
      Tsi_SortChannels(*ImageDataStart, Width * Height)
  EndSelect

  Select Compression
  
    Case #Tsi_Compression_Zlib
      CompBufferSize = ImageDataSize + 13 + (ImageDataSize / 100)
      
      *CompBuffer = AllocateMemory(CompBufferSize)
      If *CompBuffer = 0
        FreeMemory(*TsiBuffer)
        Tsi_LastError = #Tsi_Error_AllocMem
        Debug "Tsi_EncodeTsi: Error on allocating memory."
        ProcedureReturn 0
      EndIf
      
      If compress2(*CompBuffer, @CompBufferSize, *ImageDataStart, ImageDataSize, CompressionLevel) 
       
        FreeMemory(*CompBuffer)
        *TsiSize\i - SizeOf(Tsi_CompressionHeader)
        
        *TsiBuffer = ReAllocateMemory(*TsiBuffer, *TsiSize\i)
        If *TsiBuffer = 0
          FreeMemory(*TsiBuffer)
          Tsi_LastError = #Tsi_Error_AllocMem
          Debug "Tsi_EncodeTsi: Error on allocating memory."
          ProcedureReturn 0
        EndIf
        
        *Header = *TsiBuffer
        *ImageDataStart = *TsiBuffer + SizeOf(Tsi_Header)
        CopyMemory(*RawImageData, *ImageDataStart, ImageDataSize)
        
      Else
      
        ImageDataSize = CompBufferSize
        *TsiSize\i = SizeOf(Tsi_Header) + SizeOf(Tsi_CompressionHeader) + ImageDataSize
        
        *TsiBuffer = ReAllocateMemory(*TsiBuffer, *TsiSize\i)
        If *TsiBuffer = 0
          FreeMemory(*CompBuffer)
          FreeMemory(*TsiBuffer)
          Tsi_LastError = #Tsi_Error_AllocMem
          Debug "Tsi_EncodeTsi: Error on allocating memory."
          ProcedureReturn 0
        EndIf
        
        *Header = *TsiBuffer
        *CompHeader = *Header + SizeOf(Tsi_Header)
        *CompHeader\UncompSize = Width * Height * 4
        *CompHeader\CompSize = ImageDataSize
        
        *ImageDataStart = *TsiBuffer + SizeOf(Tsi_Header) + SizeOf(Tsi_CompressionHeader)
        CopyMemory(*CompBuffer, *ImageDataStart, ImageDataSize)
        FreeMemory(*CompBuffer)
        
        *Header\Flags = *Header\Flags | #Tsi_Flag_CompressedZlib
      
      EndIf
    
    Case #Tsi_Compression_Thor
      FreeMemory(*TsiBuffer)
      Tsi_LastError = #Tsi_Error_DecompressThor
      Debug "Tsi_EncodeTsi: Thorium compression chosen for image data compression, not supported."
      ProcedureReturn 0

  EndSelect
  
  ProcedureReturn *TsiBuffer

EndProcedure

CompilerEndIf

;---------------------------------------------------------
;TSI Encoder, PB Image
;
;Do not change anything from here!

CompilerIf #Tsi_Compile_PbImageEncoder = #True

;Encodes a image to raw image data suitable for encoding to a TSI.
;Returns a pointer to the raw image data.
;Returns size of the raw image data in *RawImageDataSize.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_EncodeImage2RawImageData(Image.i, *RawImageDataSize.Integer)

  Protected.i X, Y, Width, Height, DestLineLen, Pitch, PixelFormat, SrcY
  Protected *ImageBuffer
  Protected *RawImageBuffer
  Protected *SrcPixel.Long
  Protected *DestPixel.Long

  If ImageDepth(Image) <> 32
    Tsi_LastError = #Tsi_Error_PbImageWrongPixelFormat
    Debug "Tsi_EncodeImage2RawImageData: Wrong pixel format of PB image, depth is not 32bit."  
    ProcedureReturn 0
  EndIf

  Width = ImageWidth(Image)
  Height = ImageHeight(Image)
  *RawImageDataSize\i = Width * Height * 4

  *RawImageBuffer = AllocateMemory(*RawImageDataSize\i)
  If *RawImageBuffer = 0
    Tsi_LastError = #Tsi_Error_AllocMem
    Debug "Tsi_EncodeImage2RawImageData: Error on allocating memory."
    ProcedureReturn 0
  EndIf

  If StartDrawing(ImageOutput(Image)) = 0
    FreeMemory(*RawImageBuffer)
    Tsi_LastError = #Tsi_Error_PbImageAccess
    Debug "Tsi_EncodeImage2RawImageData: Can't access PB image."
    ProcedureReturn 0
  EndIf

  *ImageBuffer = DrawingBuffer()
  If *ImageBuffer = 0
    StopDrawing()
    FreeMemory(*RawImageBuffer)
    Tsi_LastError = #Tsi_Error_PbImageBufferAccess
    Debug "Tsi_EncodeImage2RawImageData: Can't access PB image buffer."
    ProcedureReturn 0
  EndIf

  Pitch = DrawingBufferPitch()
  PixelFormat = DrawingBufferPixelFormat()

  DestLineLen = Width * 4

  Select PixelFormat

    Case #PB_PixelFormat_32Bits_BGR
    
      If Pitch = DestLineLen
        CopyMemory(*ImageBuffer, *RawImageBuffer, *RawImageDataSize\i)
      Else
        For Y = 0 To Height - 1
          CopyMemory(*ImageBuffer + (Pitch * Y), *RawImageBuffer + (DestLineLen * Y), DestLineLen)
        Next
      EndIf

    Case #PB_PixelFormat_32Bits_RGB
      
      *DestPixel = *RawImageBuffer
      For Y = 0 To Height - 1
        *SrcPixel = *ImageBuffer + (Pitch * Y)
        For X = 0 To Width - 1
          *DestPixel\l = Tsi_BGRA2RGBA(*SrcPixel\l)
          *SrcPixel + 4
          *DestPixel + 4
        Next
      Next

    Case #Tsi_PB_PixelFormat_32Bits_BGR_ReversedY

      SrcY = Height - 1
      For Y = 0 To Height - 1
        CopyMemory(*ImageBuffer + (Pitch * SrcY), *RawImageBuffer + (DestLineLen * Y), DestLineLen)
        SrcY - 1
      Next

    Case #Tsi_PB_PixelFormat_32Bits_RGB_ReversedY

      *DestPixel = *RawImageBuffer
      SrcY = Height - 1
      For Y = 0 To Height - 1
        *SrcPixel = *ImageBuffer + (Pitch * SrcY)
        For X = 0 To Width - 1
          *DestPixel\l = Tsi_BGRA2RGBA(*SrcPixel\l)
          *SrcPixel + 4
          *DestPixel + 4
        Next
        SrcY - 1
      Next

    Default

      StopDrawing()
      FreeMemory(*RawImageBuffer)
      Tsi_LastError = #Tsi_Error_PbImageWrongPixelFormat
      Debug "Tsi_EncodeImage2RawImageData: Wrong pixel format of PB image."
      ProcedureReturn 0

  EndSelect

  StopDrawing()

  ProcedureReturn *RawImageBuffer

EndProcedure

;Similar to PB's SaveImage. But saves the TSI to memory and not to hard disk.
;Returns a pointer to the TSI.
;Returns size of the TSI in *TsiSize.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_Save2MemoryFromImage(Image.i, *TsiSize.Integer, Filter.i, Transformation.i, Compression.i, CompressionLevel.i)

  Protected.i RawImageSize
  Protected *RawImageBuffer
  Protected *Tsi

  *RawImageBuffer = Tsi_EncodeImage2RawImageData(Image, @RawImageSize)
  If *RawImageBuffer = 0
    ProcedureReturn 0
  EndIf

  *Tsi = Tsi_EncodeTsi(*RawImageBuffer, *TsiSize, ImageWidth(Image), ImageHeight(Image), Filter, Transformation, Compression, CompressionLevel)
  FreeMemory(*RawImageBuffer)
  
  ProcedureReturn *Tsi

EndProcedure

;Similar to PB's SaveImage. But saves the TSI to memory and not to hard disk.
;The difference to Tsi_Save2MemoryFromImage is that is will choose the best
;filter and transformation automaticly by testing which is the best for the given
;image. Therefor it's much slower but outputs allways the best optimized TSI.
;Returns a pointer to the TSI.
;Returns size of the TSI in *TsiSize.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_BestSave2MemoryFromImage(Image.i, *TsiSize.Integer)

  Protected.i Width, Height, RawImageSize, BestTsiSize, NewTsiSize, BestTransformation, BestFilter
  Protected *RawImageData
  Protected *BestTsi
  Protected *NewTsi
  
  Width  = ImageWidth(Image)
  Height = ImageHeight(Image)
  
  *RawImageData = Tsi_EncodeImage2RawImageData(Image, @RawImageSize)
  If *RawImageData = 0
    ProcedureReturn 0
  EndIf
  
  *BestTsi = Tsi_EncodeTsi(*RawImageData, @BestTsiSize, Width, Height, #Tsi_Filter_None, #Tsi_Trandformation_None, #Tsi_Compression_Zlib, 9)
  If *BestTsi = 0
    FreeMemory(*RawImageData)
    ProcedureReturn 0
  EndIf
  
  *NewTsi = Tsi_EncodeTsi(*RawImageData, @NewTsiSize, Width, Height, #Tsi_Filter_None, #Tsi_Transformation_AlphaSort, #Tsi_Compression_Zlib, 9)
  If *NewTsi = 0
    FreeMemory(*RawImageData)
    FreeMemory(*BestTsi)
    ProcedureReturn 0
  EndIf
  If NewTsiSize < BestTsiSize
    BestTransformation = #Tsi_Transformation_AlphaSort
    FreeMemory(*BestTsi)
    *BestTsi = *NewTsi
    BestTsiSize = NewTsiSize
  Else
    BestTransformation = #Tsi_Trandformation_None
    FreeMemory(*NewTsi)
  EndIf
  
  BestFilter = #Tsi_Filter_None
  
  *NewTsi = Tsi_EncodeTsi(*RawImageData, @NewTsiSize, Width, Height, #Tsi_Filter_Sub, BestTransformation, #Tsi_Compression_Zlib, 9)
  If *NewTsi = 0
    FreeMemory(*RawImageData)
    FreeMemory(*BestTsi)
    ProcedureReturn 0
  EndIf
  If NewTsiSize < BestTsiSize
    BestFilter = #Tsi_Filter_Sub
    FreeMemory(*BestTsi)
    *BestTsi = *NewTsi
    BestTsiSize = NewTsiSize
  Else
    FreeMemory(*NewTsi)
  EndIf
  
  *NewTsi = Tsi_EncodeTsi(*RawImageData, @NewTsiSize, Width, Height, #Tsi_Filter_Up, BestTransformation, #Tsi_Compression_Zlib, 9)
  If *NewTsi = 0
    FreeMemory(*RawImageData)
    FreeMemory(*BestTsi)
    ProcedureReturn 0
  EndIf
  If NewTsiSize < BestTsiSize
    BestFilter = #Tsi_Filter_Up
    FreeMemory(*BestTsi)
    *BestTsi = *NewTsi
    BestTsiSize = NewTsiSize
  Else
    FreeMemory(*NewTsi)
  EndIf
  
  *NewTsi = Tsi_EncodeTsi(*RawImageData, @NewTsiSize, Width, Height, #Tsi_Filter_Average, BestTransformation, #Tsi_Compression_Zlib, 9)
  If *NewTsi = 0
    FreeMemory(*RawImageData)
    FreeMemory(*BestTsi)
    ProcedureReturn 0
  EndIf
  If NewTsiSize < BestTsiSize
    BestFilter = #Tsi_Filter_Average
    FreeMemory(*BestTsi)
    *BestTsi = *NewTsi
    BestTsiSize = NewTsiSize
  Else
    FreeMemory(*NewTsi)
  EndIf

  FreeMemory(*RawImageData)
  *TsiSize\i = BestTsiSize
  
  ProcedureReturn *BestTsi

EndProcedure

;Similar to PB's SaveImage.
;Returns #True if it succeeds.
;Returns #False if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_SaveFromImage(Image.i, FileName.s, Filter.i, Transformation.i, Compression.i, CompressionLevel.i)

  Protected.i File, TsiSize
  Protected *Tsi
  
  *Tsi = Tsi_Save2MemoryFromImage(Image, @TsiSize, Filter, Transformation, Compression, CompressionLevel)
  If *Tsi = 0
    ProcedureReturn #False
  EndIf
  
  File = CreateFile(#PB_Any, FileName)
  If File = 0
    FreeMemory(*Tsi)
    Tsi_LastError = #Tsi_Error_CreateFile   
    Debug "Tsi_SaveFromImage: Can't create file: " + FileName
    ProcedureReturn #False
  EndIf
  
  WriteData(File, *Tsi, TsiSize)
  CloseFile(File)
  FreeMemory(*Tsi)
  
  ProcedureReturn #True
  
EndProcedure

;Similar to PB's SaveImage.
;The difference to Tsi_SaveFromImage is that is will choose the best
;filter and transformation automaticly by testing which is the best for the given
;image. Therefor it's much slower but outputs allways the best optimized TSI.
;Returns #True if it succeeds.
;Returns #False if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_BestSaveFromImage(Image.i, FileName.s)

  Protected.i File, TsiSize
  Protected *Tsi
  
  *Tsi = Tsi_BestSave2MemoryFromImage(Image, @TsiSize)
  If *Tsi = 0
    ProcedureReturn #False
  EndIf
  
  File = CreateFile(#PB_Any, FileName)
  If File = 0
    FreeMemory(*Tsi)
    Tsi_LastError = #Tsi_Error_CreateFile   
    Debug "Tsi_BestSaveFromImage: Can't create file: " + FileName
    ProcedureReturn #False
  EndIf
  
  WriteData(File, *Tsi, TsiSize)
  CloseFile(File)
  FreeMemory(*Tsi)
  
  ProcedureReturn #True

EndProcedure

CompilerEndIf

;---------------------------------------------------------
;TSI Encoder, PB Sprite
;
;Do not change anything from here!

CompilerIf #Tsi_Compile_PbSpriteEncoder = #True

;Encodes a sprite to raw image data suitable for encoding to a TSI.
;Returns a pointer to the raw image data.
;Returns size of the raw image data in *RawImageDataSize.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_EncodeSprite2RawImageData(Sprite.i, *RawImageDataSize.Integer)

  Protected.i X, Y, Width, Height, DestLineLen, RawImageSize, Pitch, PixelFormat, SrcY
  Protected *SpriteBuffer
  Protected *RawImageBuffer
  Protected *SrcPixel.Long
  Protected *DestPixel.Long

  If SpriteDepth(Sprite) <> 32
    Tsi_LastError = #Tsi_Error_PbSpriteWrongPixelFormat
    Debug "Tsi_EncodeSprite2RawImageData: Wrong pixel format of PB sprite, depth is not 32bit."  
    ProcedureReturn 0
  EndIf

  Width = SpriteWidth(Sprite)
  Height = SpriteHeight(Sprite)
  *RawImageDataSize\i = Width * Height * 4

  *RawImageBuffer = AllocateMemory(*RawImageDataSize\i)
  If *RawImageBuffer = 0
    Tsi_LastError = #Tsi_Error_AllocMem
    Debug "Tsi_EncodeSprite2RawImageData: Error on allocating memory."
    ProcedureReturn 0
  EndIf

  If StartDrawing(SpriteOutput(Sprite)) = 0
    FreeMemory(*RawImageBuffer)
    Tsi_LastError = #Tsi_Error_PbSpriteAccess
    Debug "Tsi_EncodeSprite2RawImageData: Can't access PB sprite."
    ProcedureReturn 0
  EndIf

  *SpriteBuffer = DrawingBuffer()
  If *SpriteBuffer = 0
    StopDrawing()
    FreeMemory(*RawImageBuffer)
    Tsi_LastError = #Tsi_Error_PbSpriteBufferAccess
    Debug "Tsi_EncodeSprite2RawImageData: Can't access PB sprite buffer."
    ProcedureReturn 0
  EndIf

  Pitch = DrawingBufferPitch()
  PixelFormat = DrawingBufferPixelFormat()

  DestLineLen = Width * 4

  Select PixelFormat

    Case #PB_PixelFormat_32Bits_BGR
    
      If Pitch = DestLineLen
        CopyMemory(*SpriteBuffer, *RawImageBuffer, *RawImageDataSize\i)
      Else
        For Y = 0 To Height - 1
          CopyMemory(*SpriteBuffer + (Pitch * Y), *RawImageBuffer + (DestLineLen * Y), DestLineLen)
        Next
      EndIf

    Case #PB_PixelFormat_32Bits_RGB
      
      *DestPixel = *RawImageBuffer
      For Y = 0 To Height - 1
        *SrcPixel = *SpriteBuffer + (Pitch * Y)
        For X = 0 To Width - 1
          *DestPixel\l = Tsi_BGRA2RGBA(*SrcPixel\l)
          *SrcPixel + 4
          *DestPixel + 4
        Next
      Next

    Case #Tsi_PB_PixelFormat_32Bits_BGR_ReversedY

      SrcY = Height - 1
      For Y = 0 To Height - 1
        CopyMemory(*SpriteBuffer + (Pitch * SrcY), *RawImageBuffer + (DestLineLen * Y), DestLineLen)
        SrcY - 1
      Next

    Case #Tsi_PB_PixelFormat_32Bits_RGB_ReversedY

      *DestPixel = *RawImageBuffer
      SrcY = Height - 1
      For Y = 0 To Height - 1
        *SrcPixel = *SpriteBuffer + (Pitch * SrcY)
        For X = 0 To Width - 1
          *DestPixel\l = Tsi_BGRA2RGBA(*SrcPixel\l)
          *SrcPixel + 4
          *DestPixel + 4
        Next
        SrcY - 1
      Next

    Default

      StopDrawing()
      FreeMemory(*RawImageBuffer)
      Tsi_LastError = #Tsi_Error_PbSpriteWrongPixelFormat
      Debug "Tsi_EncodeSprite2RawImageData: Wrong pixel format of PB sprite."
      ProcedureReturn 0

  EndSelect

  StopDrawing()

  ProcedureReturn *RawImageBuffer

EndProcedure

;Similar to PB's SaveSprite. But saves the TSI to memory and not to hard disk.
;Returns a pointer to the TSI.
;Returns size of the TSI in *TsiSize.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_Save2MemoryFromSprite(Sprite.i, *TsiSize.Integer, Filter.i, Transformation.i, Compression.i, CompressionLevel.i)

  Protected.i RawImageSize
  Protected *RawImageBuffer
  Protected *Tsi

  *RawImageBuffer = Tsi_EncodeSprite2RawImageData(Sprite, @RawImageSize)
  If *RawImageBuffer = 0
    ProcedureReturn 0
  EndIf

  *Tsi = Tsi_EncodeTsi(*RawImageBuffer, *TsiSize, SpriteWidth(Sprite), SpriteHeight(Sprite), Filter, Transformation, Compression, CompressionLevel)
  FreeMemory(*RawImageBuffer)
  
  ProcedureReturn *Tsi

EndProcedure

;Similar to PB's SaveSprite. But saves the TSI to memory and not to hard disk.
;The difference to Tsi_Save2MemoryFromSprite is that is will choose the best
;filter and transformation automaticly by testing which is the best for the given
;sprite. Therefor it's much slower but outputs allways the best optimized TSI.
;Returns a pointer to the TSI.
;Returns size of the TSI in *TsiSize.
;Returns 0 if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_BestSave2MemoryFromSprite(Sprite.i, *TsiSize.Integer)

  Protected.i Width, Height, RawImageSize, BestTsiSize, NewTsiSize, BestTransformation, BestFilter
  Protected *RawImageData
  Protected *BestTsi
  Protected *NewTsi
  
  Width  = SpriteWidth(Sprite)
  Height = SpriteHeight(Sprite)
  
  *RawImageData = Tsi_EncodeSprite2RawImageData(Sprite, @RawImageSize)
  If *RawImageData = 0
    ProcedureReturn 0
  EndIf
  
  *BestTsi = Tsi_EncodeTsi(*RawImageData, @BestTsiSize, Width, Height, #Tsi_Filter_None, #Tsi_Trandformation_None, #Tsi_Compression_Zlib, 9)
  If *BestTsi = 0
    FreeMemory(*RawImageData)
    ProcedureReturn 0
  EndIf
  
  *NewTsi = Tsi_EncodeTsi(*RawImageData, @NewTsiSize, Width, Height, #Tsi_Filter_None, #Tsi_Transformation_AlphaSort, #Tsi_Compression_Zlib, 9)
  If *NewTsi = 0
    FreeMemory(*RawImageData)
    FreeMemory(*BestTsi)
    ProcedureReturn 0
  EndIf
  If NewTsiSize < BestTsiSize
    BestTransformation = #Tsi_Transformation_AlphaSort
    FreeMemory(*BestTsi)
    *BestTsi = *NewTsi
    BestTsiSize = NewTsiSize
  Else
    BestTransformation = #Tsi_Trandformation_None
    FreeMemory(*NewTsi)
  EndIf
  
  BestFilter = #Tsi_Filter_None
  
  *NewTsi = Tsi_EncodeTsi(*RawImageData, @NewTsiSize, Width, Height, #Tsi_Filter_Sub, BestTransformation, #Tsi_Compression_Zlib, 9)
  If *NewTsi = 0
    FreeMemory(*RawImageData)
    FreeMemory(*BestTsi)
    ProcedureReturn 0
  EndIf
  If NewTsiSize < BestTsiSize
    BestFilter = #Tsi_Filter_Sub
    FreeMemory(*BestTsi)
    *BestTsi = *NewTsi
    BestTsiSize = NewTsiSize
  Else
    FreeMemory(*NewTsi)
  EndIf
  
  *NewTsi = Tsi_EncodeTsi(*RawImageData, @NewTsiSize, Width, Height, #Tsi_Filter_Up, BestTransformation, #Tsi_Compression_Zlib, 9)
  If *NewTsi = 0
    FreeMemory(*RawImageData)
    FreeMemory(*BestTsi)
    ProcedureReturn 0
  EndIf
  If NewTsiSize < BestTsiSize
    BestFilter = #Tsi_Filter_Up
    FreeMemory(*BestTsi)
    *BestTsi = *NewTsi
    BestTsiSize = NewTsiSize
  Else
    FreeMemory(*NewTsi)
  EndIf
  
  *NewTsi = Tsi_EncodeTsi(*RawImageData, @NewTsiSize, Width, Height, #Tsi_Filter_Average, BestTransformation, #Tsi_Compression_Zlib, 9)
  If *NewTsi = 0
    FreeMemory(*RawImageData)
    FreeMemory(*BestTsi)
    ProcedureReturn 0
  EndIf
  If NewTsiSize < BestTsiSize
    BestFilter = #Tsi_Filter_Average
    FreeMemory(*BestTsi)
    *BestTsi = *NewTsi
    BestTsiSize = NewTsiSize
  Else
    FreeMemory(*NewTsi)
  EndIf

  FreeMemory(*RawImageData)
  *TsiSize\i = BestTsiSize
  
  ProcedureReturn *BestTsi

EndProcedure

;Similar to PB's SaveSprite.
;Returns #True if it succeeds.
;Returns #False if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_SaveFromSprite(Sprite.i, FileName.s, Filter.i, Transformation.i, Compression.i, CompressionLevel.i)

  Protected.i File, TsiSize
  Protected *Tsi
  
  *Tsi = Tsi_Save2MemoryFromSprite(Sprite, @TsiSize, Filter, Transformation, Compression, CompressionLevel)
  If *Tsi = 0
    ProcedureReturn #False
  EndIf
  
  File = CreateFile(#PB_Any, FileName)
  If File = 0
    FreeMemory(*Tsi)
    Tsi_LastError = #Tsi_Error_CreateFile   
    Debug "Tsi_SaveFromSprite: Can't create file: " + FileName
    ProcedureReturn #False
  EndIf
  
  WriteData(File, *Tsi, TsiSize)
  CloseFile(File)
  FreeMemory(*Tsi)
  
  ProcedureReturn #True
  
EndProcedure

;Similar to PB's SaveSprite.
;The difference to Tsi_SaveFromSprite is that is will choose the best
;filter and transformation automaticly by testing which is the best for the given
;image. Therefor it's much slower but outputs allways the best optimized TSI.
;Returns #True if it succeeds.
;Returns #False if it fails, check Tsi_LastError for error code.
Procedure.i Tsi_BestSaveFromSprite(Sprite.i, FileName.s)

  Protected.i File, TsiSize
  Protected *Tsi
  
  *Tsi = Tsi_BestSave2MemoryFromSprite(Sprite, @TsiSize)
  If *Tsi = 0
    ProcedureReturn #False
  EndIf
  
  File = CreateFile(#PB_Any, FileName)
  If File = 0
    FreeMemory(*Tsi)
    Tsi_LastError = #Tsi_Error_CreateFile   
    Debug "Tsi_BestSaveFromSprite: Can't create file: " + FileName
    ProcedureReturn #False
  EndIf
  
  WriteData(File, *Tsi, TsiSize)
  CloseFile(File)
  FreeMemory(*Tsi)
  
  ProcedureReturn #True

EndProcedure

CompilerEndIf

; IDE Options = PureBasic 4.40 Beta 5 (Windows - x86)
; CursorPosition = 27
; Folding = AAAAw
; EnableXP
; DisableDebugger